#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Implements error-handling according to XCP spec.
"""

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from collections import namedtuple
import inspect
import functools
from pprint import pprint

import sys
import os

import time
import threading
import types

from pyxcp.types import XcpResponseError, XcpTimeoutError, XcpError, Command, COMMAND_CATEGORIES
from pyxcp.errormatrix import ERROR_MATRIX, PreAction, Action
from ..logger import Logger


class SingletonBase(object):
    _lock = threading.Lock()

    def __new__(cls, *args, **kws):
        # Double-Checked Locking
        if not hasattr(cls, "_instance"):
            try:
                cls._lock.acquire()
                if not hasattr(cls, "_instance"):
                    cls._instance = super(SingletonBase, cls).__new__(cls)
            finally:
                cls._lock.release()
        return cls._instance


Function = namedtuple("Function", "fun arguments")  # store: var | load: var


class InternalError(Exception):
    """Indicates an internal error, like invalid service."""


class UnhandledError(Exception):
    """"""


class UnrecoverableError(Exception):
    """"""


def func_name(func):
    return func.__qualname__ if func is not None else None


def getErrorHandler(service):
    """"""
    return ERROR_MATRIX.get(service)


def getTimeoutHandler(service):
    """"""
    handler = getErrorHandler(service)
    if handler is None:
        raise InternalError("Invalid Service")
    return handler.get(XcpError.ERR_TIMEOUT)


def getActions(service, error_code):
    """"""
    error_str = str(error_code)
    if error_code == XcpError.ERR_TIMEOUT:
        preActions, actions = getTimeoutHandler(service)
    else:
        eh = getErrorHandler(service)
        if eh is None:
            raise InternalError("Invalid Service 0x{:02x}".format(service))
        print("Try to handle error -- Service: {} Error-Code: {}".format(service.name, error_code))
        handler = eh.get(error_str)
        if handler is None:
            raise UnhandledError("Service '{}' has no handler for '{}'.".format(service.name, error_code))
        preActions, actions = handler
    return preActions, actions


def actionIter(actions):
    """Iterate over action from :file:`errormatrix.py`"""
    if isinstance(actions, (tuple, list)):
        for item in actions:
            yield item
    else:
        yield actions


class Arguments:
    """Container for positional and keyword arguments.

    Parameters
    ----------
        args: tuple
            Positional arguments
        kwargs: dict
            Keyword arguments.
    """

    def __init__(self, args=None, kwargs={}):
        if args is None:
            self.args = []
        else:
            if not hasattr(args, "__iter__"):
                self.args = (args,)
            else:
                self.args = tuple(args)
        self.kwargs = kwargs

    def __str__(self):
        res = "{}(ARGS = {}, KWS = {})".format(self.__class__.__name__, self.args, self.kwargs)
        return res

    def __eq__(self, other):
        return (self.args == other.args if other is not None else ()) and (
            self.kwargs == other.kwargs if other is not None else {}
        )

    __repr__ = __str__


class Repeater:
    """A required action of some XCP errorhandler is repetition.

    Parameters
    ----------
        initial_value: int
            The actual values are predetermined by XCP:
                - REPEAT (one time)
                - REPEAT_2_TIMES (two times)
                - REPEAT_INF_TIMES ("forever")
    """

    REPEAT = 1
    REPEAT_2_TIMES = 2
    INFINITE = -1

    def __init__(self, initial_value: int):
        self._counter = initial_value

    def repeat(self):
        """Check if repetition is required.

        Returns
        -------
            bool
        """
        # print("\t\tCOUNTER:", self._counter)
        if self._counter == Repeater.INFINITE:
            return True
        elif self._counter > 0:
            self._counter -= 1
            return True
        else:
            return False


def display_error():
    """Display error information.

    TODO: callback.

    """


class Handler:
    """"""

    logger = Logger(__name__)

    def __init__(self, instance, func, arguments, error_code=None):
        self.instance = instance
        if hasattr(func, "__closure__") and func.__closure__:
            self.func = func.__closure__[0].cell_contents  # Use original, undecorated function to prevent
            # nasty recursion problems.
        else:
            self.func = func
        self.arguments = arguments
        self.service = self.instance.service
        self.error_code = error_code

    def __str__(self):
        return "Handler(func = {} arguments = {} service = {} error_code = {})".format(
            func_name(self.func), self.arguments, self.service, self.error_code
        )

    def __eq__(self, other):
        return (self.instance == other.instance) and (self.func == other.func) and (self.arguments == other.arguments)

    def execute(self):
        self.logger.debug("EXECUTE func = {} arguments = {})".format(func_name(self.func), self.arguments))
        if isinstance(self.func, types.MethodType):
            return self.func(*self.arguments.args, **self.arguments.kwargs)
        else:
            return self.func(self.instance, *self.arguments.args, **self.arguments.kwargs)

    def actions(self, preActions, actions):
        """Preprocess errorhandling pre-actions and actions."""
        result_pre_actions = []
        result_actions = []
        repetitionCount = 0
        for item in actionIter(preActions):
            if item == PreAction.NONE:
                pass
            elif item == PreAction.WAIT_T7:
                time.sleep(0.02)  # Completely arbitrary for now.
            elif item == PreAction.SYNCH:
                fn = Function(self.instance.synch, Arguments())
                result_pre_actions.append(fn)
            elif item == PreAction.GET_SEED_UNLOCK:
                raise NotImplementedError("GET_SEED_UNLOCK")
            elif item == PreAction.SET_MTA:
                fn = Function(self.instance.setMta, Arguments(self.instance.mta))
                result_pre_actions.append(fn)
            elif item == PreAction.SET_DAQ_PTR:
                fn = Function(self.instance.setDaqPtr, Arguments(self.instance.currentDaqPtr))
            elif item == PreAction.START_STOP_X:
                raise NotImplementedError("START_STOP_X")
            elif item == PreAction.REINIT_DAQ:
                raise NotImplementedError("REINIT_DAQ")
            elif item == PreAction.DISPLAY_ERROR:
                pass
            elif item == PreAction.DOWNLOAD:
                raise NotImplementedError("DOWNLOAD")
            elif item == PreAction.PROGRAM:
                raise NotImplementedError("PROGRAM")
            elif item == PreAction.UPLOAD:
                raise NotImplementedError("UPLOAD")
            elif item == PreAction.UNLOCK_SLAVE:
                resource = COMMAND_CATEGORIES.get(self.instance.service)
                raise NotImplementedError("UNLOCK_SLAVE")
        for item in actionIter(actions):
            if item == Action.NONE:
                pass
            elif item == Action.DISPLAY_ERROR:
                raise UnhandledError("Could not proceed due to unhandled error.")
            elif item == Action.RETRY_SYNTAX:
                raise UnhandledError("Could not proceed due to unhandled error.")
            elif item == Action.RETRY_PARAM:
                raise UnhandledError("Could not proceed due to unhandled error.")
            elif item == Action.USE_A2L:
                raise UnhandledError("Could not proceed due to unhandled error.")
            elif item == Action.USE_ALTERATIVE:
                raise UnhandledError("Could not proceed due to unhandled error.")  # TODO: check alternatives.
            elif item == Action.REPEAT:
                repetitionCount = Repeater.REPEAT
            elif item == Action.REPEAT_2_TIMES:
                repetitionCount = Repeater.REPEAT_2_TIMES
            elif item == Action.REPEAT_INF_TIMES:
                repetitionCount = Repeater.INFINITE
            elif item == Action.RESTART_SESSION:
                raise UnhandledError("Could not proceed due to unhandled error.")
            elif item == Action.TERMINATE_SESSION:
                raise UnhandledError("Could not proceed due to unhandled error.")
            elif item == Action.SKIP:
                pass
            elif item == Action.NEW_FLASH_WARE:
                raise UnhandledError("Could not proceed due to unhandled error")
        return result_pre_actions, result_actions, Repeater(repetitionCount)


class HandlerStack:
    """"""

    def __init__(self):
        self._stack = []

    def push(self, handler):
        self._stack.append(handler)

    def pop(self):
        if len(self) > 0:
            res = self._stack.pop()

    def tos(self):
        if len(self) > 0:
            return self._stack[-1]
        else:
            return None

    def empty(self):
        return self._stack == []

    def __len__(self):
        return len(self._stack)

    def __repr__(self):
        result = []
        for idx in range(len(self)):
            result.append(str(self[idx]))
        return "{}".format("\n".join(result))

    def __getitem__(self, ndx):
        return self._stack[ndx]

    __str__ = __repr__


class Executor(SingletonBase):
    """"""

    handlerStack = HandlerStack()
    repeater = None
    newHandler = True
    logger = Logger(__name__)
    previous_error_code = None
    error_code = None
    func = None
    arguments = None

    def __call__(self, inst, func, arguments):
        self.logger.debug("__call__({})".format(func.__qualname__))
        self.inst = inst
        self.func = func
        self.arguments = arguments
        handler = Handler(inst, func, arguments)
        self.handlerStack.push(handler)
        while True:
            try:
                handler = self.handlerStack.tos()
                res = handler.execute()
            except XcpResponseError as e:
                self.logger.error("XcpResponseError [{}]".format(str(e)))
                if self.newHandler:
                    self.error_code = e.get_error_code()
            except XcpTimeoutError as e:
                self.logger.error("XcpTimeoutError")
                if self.newHandler:
                    self.error_code = XcpError.ERR_TIMEOUT
            except Exception as e:
                raise UnrecoverableError("Don't know how to handle exception '{}'".format(repr(e))) from e
            else:
                self.error_code = None
                # print("\t\t\t*** SUCCESS ***")
                self.handlerStack.pop()
                if self.handlerStack.empty():
                    # print("OK, all handlers passed: '{}'.".format(res))
                    return res
            if self.error_code != None:
                preActions, actions, repeater = handler.actions(*getActions(inst.service, self.error_code))
                self.repeater = repeater
                for f, a in reversed(preActions):
                    self.handlerStack.push(Handler(inst, f, a, self.error_code))
                self.newHandler = False
                continue
            self.previous_error_code = self.error_code
            if self.repeater:
                if self.repeater.repeat():
                    continue
                else:
                    raise UnrecoverableError("Finish for now.")


def wrapped(func):
    """This decorator is XCP error-handling enabled."""

    @functools.wraps(func)
    def inner(*args, **kwargs):
        handle_errors = os.environ.get("PYXCP_HANDLE_ERRORS", True)
        if handle_errors in (False, "False", "false"):
            handle_errors = False
        if handle_errors:
            inst = args[0]  # First parameter is 'self'.
            arguments = Arguments(args[1:], kwargs)
            executor = Executor()
            res = executor(inst, func, arguments)
        else:
            res = func(*args, **kwargs)
        return res

    return inner



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Parse (transport-layer specific) command line parameters
and create a XCP master instance.
"""

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import argparse

from pyxcp.config import readConfiguration
from pyxcp.master import Master
from pyxcp.transport.can import try_to_install_system_supplied_drivers, registered_drivers

try_to_install_system_supplied_drivers()

CAN_DRIVERS = registered_drivers()


class ArgumentParser:
    """

    Parameter
    ---------
    callout: callable
        Process user-supplied arguments.
    """

    def __init__(self, callout=None, *args, **kws):
        self.callout = callout
        kws.update(formatter_class=argparse.RawDescriptionHelpFormatter, add_help=True)
        self._parser = argparse.ArgumentParser(*args, **kws)
        self._parser.add_argument(
            "-c",
            "--config-file",
            type=argparse.FileType("r"),
            dest="conf",
            help="File to read (extended) parameters from.",
        )
        self._parser.add_argument("-l", "--loglevel", choices=["ERROR", "WARN", "INFO", "DEBUG"], default="INFO")
        self._parser.epilog = "To get specific help on transport layers\nuse <layer> -h, e.g. {} eth -h".format(
            self._parser.prog
        )
        self._args = []

    @property
    def args(self):
        return self._args

    def run(self):
        """"""
        self._args = self.parser.parse_args()
        args = self.args
        config = readConfiguration(args.conf)
        config["LOGLEVEL"] = args.loglevel
        if not "TRANSPORT" in config:
            raise AttributeError("TRANSPORT must be specified in config!")
        transport = config["TRANSPORT"].lower()
        master = Master(transport, config=config)
        if self.callout:
            self.callout(master, args)
        return master

    @property
    def parser(self):
        return self._parser



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
python-can driver for Ixxat interfaces.
"""

__copyright__ = """
    pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2009-2020 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from can import Bus, CanError, Message

import pyxcp.transport.can as can
import pyxcp.transport.candriver.python_can as python_can


class Ixxat(python_can.PythonCAN, can.CanInterfaceBase):
    """"""

    PARAMETER_MAP = {
        #                           Type    Req'd   Default
        "UNIQUE_HARDWARE_ID": (str, False, None),
        "RX_FIFO_SIZE": (int, False, 16),
        "TX_FIFO_SIZE": (int, False, 16),
    }

    PARAMETER_TO_KW_ARG_MAP = {
        "UNIQUE_HARDWARE_ID": "UniqueHardwareId",
        "RX_FIFO_SIZE": "rxFifoSize",
        "TX_FIFO_SIZE": "txFifoSize",
    }

    def __init__(self):
        super(Ixxat, self).__init__(bustype="ixxat")



import pytest

from pyxcp.transport.can import (
    setDLC,
    calculateFilter,
    CAN_EXTENDED_ID,
    isExtendedIdentifier,
    stripIdentifier,
    Identifier,
    MAX_11_BIT_IDENTIFIER,
    MAX_29_BIT_IDENTIFIER,
    IdentifierOutOfRangeError,
)


def testSet0():
    assert setDLC(0) == 0


def testSet4():
    assert setDLC(4) == 4


def testSet8():
    assert setDLC(8) == 8


def testSet9():
    assert setDLC(9) == 12


def testSet12():
    assert setDLC(12) == 12


def testSet13():
    assert setDLC(13) == 16


def testSet16():
    assert setDLC(16) == 16


def testSet17():
    assert setDLC(17) == 20


def testSet20():
    assert setDLC(20) == 20


def testSet23():
    assert setDLC(23) == 24


def testSet24():
    assert setDLC(24) == 24


def testSet25():
    assert setDLC(25) == 32


def testSet32():
    assert setDLC(32) == 32


def testSet33():
    assert setDLC(33) == 48


def testSet48():
    assert setDLC(48) == 48


def testSet49():
    assert setDLC(49) == 64


def testSet64():
    assert setDLC(64) == 64


def testSet128():
    with pytest.raises(ValueError):
        setDLC(128)


def testNegative():
    with pytest.raises(ValueError):
        setDLC(-1)


def testfilter1():
    assert calculateFilter([0x101, 0x102, 0x103]) == (0x100, 0x7FC)


def testfilter2():
    assert calculateFilter([0x101, 0x102 | CAN_EXTENDED_ID, 0x103]) == (0x100, 0x1FFFFFFC)


def testfilter3():
    assert calculateFilter([0x1567 | CAN_EXTENDED_ID]) == (0x1567, 0x1FFFFFFF)


def testfilter4():
    assert (
        calculateFilter(
            [
                0x1560 | CAN_EXTENDED_ID,
                0x1561,
                0x1562,
                0x1563,
                0x1564,
                0x1565,
                0x1566,
                0x1567,
                0x1568,
                0x1569,
                0x156A,
                0x156B,
                0x156C,
                0x156D,
                0x1563,
                0x156F,
            ]
        )
        == (0x1560, 0x1FFFFFF0)
    )


def testfilter5():
    assert calculateFilter([0x1560 | CAN_EXTENDED_ID, 0x1561, 0x1562, 0x1563, 0x1564, 0x1565, 0x1566, 0x1567]) == (
        0x1560,
        0x1FFFFFF8,
    )


def testIsExtendedIdentifier1():
    assert isExtendedIdentifier(0x280) is False


def testIsExtendedIdentifier2():
    assert isExtendedIdentifier(0x280 | CAN_EXTENDED_ID) is True


def testStripIdentifier1():
    assert stripIdentifier(0x280) == 0x280


def testStripIdentifier2():
    assert stripIdentifier(0x280 | CAN_EXTENDED_ID) == 0x280


# def testSamplePointToTsegs1():
#     samplePointToTsegs


def test_identifier_works1():
    i = Identifier(CAN_EXTENDED_ID | 4711)
    assert i.id == 4711
    assert i.is_extended is True
    assert i.raw_id == CAN_EXTENDED_ID | 4711


def test_identifier_works2():
    i = Identifier(111)
    assert i.id == 111
    assert i.is_extended is False
    assert i.raw_id == 111


def test_identifier_max_works1():
    i = Identifier(CAN_EXTENDED_ID | MAX_29_BIT_IDENTIFIER)
    assert i.id == MAX_29_BIT_IDENTIFIER
    assert i.is_extended is True
    assert i.raw_id == CAN_EXTENDED_ID | MAX_29_BIT_IDENTIFIER


def test_identifier_max_works2():
    i = Identifier(MAX_11_BIT_IDENTIFIER)
    assert i.id == MAX_11_BIT_IDENTIFIER
    assert i.is_extended is False
    assert i.raw_id == MAX_11_BIT_IDENTIFIER


def test_identifier_outof_range_raises1():
    with pytest.raises(IdentifierOutOfRangeError):
        Identifier(CAN_EXTENDED_ID | (MAX_29_BIT_IDENTIFIER + 10))


def test_identifier_outof_range_raises2():
    with pytest.raises(IdentifierOutOfRangeError):
        Identifier(MAX_11_BIT_IDENTIFIER + 10)


def test_identifier_str_works1(capsys):
    i = Identifier(101)
    print(str(i), end="")
    captured = capsys.readouterr()
    assert captured.out == "Identifier(id = 0x00000065, is_extended = False)"


def test_identifier_repr_works1(capsys):
    i = Identifier(101)
    print(repr(i), end="")
    captured = capsys.readouterr()
    assert captured.out == "Identifier(0x00000065)"


def test_identifier_str_works2(capsys):
    i = Identifier(101 | CAN_EXTENDED_ID)
    print(str(i), end="")
    captured = capsys.readouterr()
    assert captured.out == "Identifier(id = 0x00000065, is_extended = True)"


def test_identifier_repr_works2(capsys):
    i = Identifier(101 | CAN_EXTENDED_ID)
    print(repr(i), end="")
    captured = capsys.readouterr()
    assert captured.out == "Identifier(0x80000065)"


def test_identifier_repr_does_the_trick1(capsys):
    i = Identifier(101)
    assert eval(repr(i)) == Identifier(101)


def test_identifier_repr_does_the_trick2(capsys):
    i = Identifier(101 | CAN_EXTENDED_ID)
    assert eval(repr(i)) == Identifier(101 | CAN_EXTENDED_ID)
